/*
   vt100.c
   Copyright 2017 Greg Sydney-Smith
   
   2017-11-11 gss. Minimal vt100 emulation for z80sim.
      With some input/examples from https://github.com/adoxa/ansicon.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are met:

   1. Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation
      and/or other materials provided with the distribution.

   THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
   INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
   FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
   DEVELOPERS AND CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
   OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
   OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
   ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/


#include <stdio.h>
#include <windows.h>
#include <conio.h>
#include "vt100.h"
#include "simfun.h"

void user_int(int), quit_int(int), term_int(int); // in simint.c


#define NROWS 24

#define FOREGROUND_BLACK 0
#define FOREGROUND_WHITE FOREGROUND_RED|FOREGROUND_GREEN|FOREGROUND_BLUE

#define BACKGROUND_BLACK 0
#define BACKGROUND_WHITE BACKGROUND_RED|BACKGROUND_GREEN|BACKGROUND_BLUE

#define is_digit(c) ('0' <= (c) && (c) <= '9')


static int dbg = 0;
static int lfdelay = 50;        // milliseconds

HANDLE hConOut;                 // handle to CONOUT$
static int state= -1;
static int nnn1, nnn2;

// screen/windows coords
static int base_y= 0;
static int cx, cy; // cursor x, y (0..79) (0..23)
static int wCols;  // how wide the window really is (might be >80)
static int bRows;
static int cAttrib=0;


int vt100_config(char *name, char *val) {
    int i;
    char *names[] = { "debug", "lfdelay", NULL };

    if (val==NULL) val="0";
    for (i=0; names[i]!=NULL && stricmp(name,names[i])!=0; i++) ;
    switch (i) {
    case 0: dbg= atoi(val); break;
    case 1: lfdelay= atoi(val); break;
    }
    return 0;
}
int vt100_init(void) {
    return 0;
}

//void saystate(char *msg, int a1, int a2) {
static void saystate(void) {
    CONSOLE_SCREEN_BUFFER_INFO info;

    GetConsoleScreenBufferInfo(hConOut,&info);
    fprintf(stderr,"curs_x=%d\n",info.dwCursorPosition.X);
    fprintf(stderr,"curs_y=%d\n",info.dwCursorPosition.Y);
    fprintf(stderr,"cols=%d\n",info.dwSize.X);
    fprintf(stderr,"rows=%d\n",info.dwSize.Y);
    fprintf(stderr,"win.L=%d\n",info.srWindow.Left);
    fprintf(stderr,"win.R=%d\n",info.srWindow.Right);
    fprintf(stderr,"win.T=%d\n",info.srWindow.Top);
    fprintf(stderr,"win.B=%d\n",info.srWindow.Bottom);
    fprintf(stderr,"\n");
}

static void getCurs(void) {
    int x, y;
    CONSOLE_SCREEN_BUFFER_INFO info;

    if (dbg>1) saystate();
    GetConsoleScreenBufferInfo(hConOut,&info);
    x= info.dwCursorPosition.X;
    y= info.dwCursorPosition.Y;
    wCols= info.dwSize.X;
    bRows= info.dwSize.Y; // buffer rows, typically 300
    if (y > base_y+NROWS) base_y= y-NROWS; // we scrolled, screen is last 24 lines

    cx= x;
    cy= y - base_y;
}


// our vt100 screen is just 24 lines of the Win console screen buffer
// Windows cursor can occupy any line in the buffer (eg 0..299)
// assumes base_y has been set so that the cursor is on screen
//   0<= (curs_y-base_y) < 24   and   0<=x<80
static void WinCursTo(int x, int y) {
    COORD Pos;

    Pos.X = x;        // 0 based
    Pos.Y = base_y + y;
    SetConsoleCursorPosition(hConOut,Pos);
}

static void FillBlank(int x, int y, int len, int attrib) {
    DWORD n;
    COORD Pos;

    Pos.X= x;
    Pos.Y= base_y + y;
    FillConsoleOutputCharacter( hConOut, ' ', len, Pos, &n);
    FillConsoleOutputAttribute( hConOut, attrib, len, Pos, &n);
}

static void vt100Mode(int n) {

    switch (n) {
    case  0: cAttrib = FOREGROUND_WHITE | BACKGROUND_BLACK; break;
    case  7: cAttrib = FOREGROUND_BLACK | BACKGROUND_WHITE; break;
    default: fprintf(stderr,"vt100Mode %d\n",n); return;
    }
    SetConsoleTextAttribute( hConOut, cAttrib);
}

static void vt100EraseScreen(int n) {
    WORD attrib = FOREGROUND_WHITE | BACKGROUND_BLACK;

    switch (n) {
    case  2: // ESC[2J Clear screen
             FillBlank(0,0,NROWS*wCols,attrib); //info.wAttributes);
             WinCursTo(0,0); break;
    default: fprintf(stderr,"vt100EraseScreen %d\n",n);
    }
}

static void vt100EraseLine(int n) { // ESC [ n K
    WORD attrib = FOREGROUND_WHITE | BACKGROUND_BLACK;

    switch (n) {
    case  0: // ESC[0K Clear to end of line
             FillBlank(cx,cy,wCols-cx,attrib); break;
    case  1: // ESC[1K Clear from start of line to cursor
             FillBlank(0,cy,cx+1,attrib); break;
    case  2: // ESC[2K Clear whole line.
             FillBlank(0,cy,wCols,attrib); break;
    }
}

static void vt100CursTo(int y, int x) {
    if (x==0) x=1;      // default if omitted
    if (y==0) y=1;
    WinCursTo(x-1,y-1); // 0 based
}


/*
   ESC [ (num (; num)) char

num is optional. 0 is assumed if omitted.
ESC [ num m   set mode
ESC [ num J   erase (0 to end screen, 1 to start screen, 2=all)
ESC [ num K   erase (0 to end line, 1 to start line, 2=all line)
ESC [ num X   erase chars 0=1 1=1 2=2 ...
ESC [ num L   insert blank line(s) 0=1 1=1 2=2 ...
ESC [ num M   delete line(s) 0=1 1=1 2=2 ...
ESC [ num P   delete character(s) 0=1 etc
ESC [ num @   insert blank character/s 0=...
[#b	repeat the previous character # times
[#A	move cursor up # lines
[#B	move cursor down # lines
[#C	move cursor right # characters
[#D	move cursor left # characters
[#I	move cursor forward # tabs
[#Z	move cursor back # tabs
[#k	move cursor up # lines
[#e	move cursor down # lines
[#a	move cursor right # characters
[#j	move cursor left # characters
[#E	move cursor down # lines and to first column
[#F	move cursor up # lines and to first column
[#G	move cursor to column #
[#`	move cursor to column #
[#d	move cursor to line #
[H	move cursor to top-left
[#H	move cursor to line # and first column
[#;#H	move cursor to line #, column #
[f	move cursor to top-left
[#f	move cursor to line # and first column
[#;#f	move cursor to line #, column #
[s	save cursor position
[u	move cursor to saved position (or top-left, if nothing was saved)
*/

static void doSeq(int n1, int n2, int action) {
    if (dbg) fprintf(stderr," ESC [ %d;%d %c\n",n1,n2,action);
    getCurs();
    switch (action) {
    case 'm' : vt100Mode(n1); break;
    case 'J' : vt100EraseScreen(n1); break;
    case 'K' : vt100EraseLine(n1); break;
    case 'H' : vt100CursTo(n1,n2); break;
    default  : printf("vt100 doSeq: unimplemented sequence (%c)\n",action);
    }
}

static void vt100out(int co) {

    if (state < 0) {
        hConOut = GetStdHandle(STD_OUTPUT_HANDLE);
        if (hConOut == INVALID_HANDLE_VALUE)
            printf("vt100out: Can't get stdout handle\n");
        state=0;
    }

    if (state==0 && co==0x1B    ) { state=1; return; } // ESC
    if (state==1 && co=='['     ) { state=2; nnn1=0; nnn2=0; return; }
    if (state==2 && is_digit(co)) { nnn1*=10; nnn1+=(co-'0'); return; } 
    if (state==2 && co==';'     ) { state=3; return; } 
    if (state==2                ) { doSeq(nnn1,nnn2,co); state=0; return; } 
    if (state==3 && is_digit(co)) { nnn2*=10; nnn2+=(co-'0'); return; } 
    if (state==3 && co==';'     ) { state=0; return; } 
    if (state==3                ) { doSeq(nnn1,nnn2,co); state=0; return; } 
    state=0;
    putch(co&0x7F);
    fflush(stdout);

    if (dbg) {
        fprintf(stderr,(' '<=co && co<='~')?"%c":"(%02X)",co);
        if (co=='\n') fprintf(stderr,"\n");
    }
    // This is needed to give Win a chance to update buffer before WM_PAINT.
    // It's an issue at bottom of buffer when need to move 299 lines up 1.
    // Better to freeze screen, update, refresh (if Win has that); but this works.
    // (Side effect of this approach is a better match to the 1979 real world)
    if (co=='\n') os_sleep(lfdelay);
}

/*
 *	I/O function port 1 write:
 *	Write byte to stdout and flush the output.
 */

void vt100_data_out(BYTE data) {
    vt100out(data);
    fflush(stdout);
}


BYTE vt100_data_in(void) {
    int x;

    if ((x=getch()) == 0) { // A Windows function key has been pressed
        x=getch(); // 2nd byte of sequence
        if (x==0x5C) quit_int(1); // SHIFT-F9  = back to simulator
        if (x==0x5D) term_int(1); // SHIFT-F10 = back to Windows
        return ' ';
    }
    return x;
}

#define TBE 0x80
#define RDA 0x40

BYTE vt100_stat_in(void) {
//    return (kbhit()?RDA:0) | TBE;
    return (kbhit()?RDA:0) | TBE | 0x04; // cromix checks UART is connected to terminal
}

// 1sec pulse for hibernate
void vt100_tick(void) {
}

